/*
 * This program returns the buffer sizes (underflow/overflow)
 * required for the given stream. The stream data is encoded
 * as produced by the scanvbr code. To compute the overflow
 * buffer requirements, this module simulates the client 
 * buffer during transmission.
 *
 * Juan Miguel del Rosario, (c) 1995.
 * All rights reserved.
 */
 
#include <stdio.h>
#include <string.h>

#define FPS 30

/*---- Globals -----*/ 
  struct Queue{
    int fsize;
    int fno;
    struct Queue *next;
  };

  struct Queue *Consumed = NULL;
  struct Queue *Current = NULL;
  struct Queue *Qhead = NULL;
  int ConsumedFrames = 0;
/*-------------------*/

double computeOflow();
double getOverflow();

/*----------< MAIN >----------------*/

#ifdef NOTSELFRUN
stat(m, filename, fnameout)
   int m;
   char *filename;
   char *fnameout;
#else
main(argc, argv) 
   int argc;
   char *argv[];
#endif
{
  int cnt;
  FILE *fd, *fdo;
  int maxOflow, maxUflow;
  int maxOflowSlots, maxUflowSlots;
  int fnum, fsize;
  int framecount;
  int delta;
  int OflowSlots, UflowSlots;
  int i, buffsize;
  double rate, time;
  double timeslot, rem;

  maxOflow = maxUflow = 0;
  maxOflowSlots = maxUflowSlots = 0;
  OflowSlots = UflowSlots = 0;
  framecount = 0;
  i = delta = 0;
  timeslot = 0;
  rem = 0;

#ifdef NOTSELFRUN
  fd = fopen(filename, "r");
  fdo = fopen(fnameout, "a+");
#else
  if (argc < 2) {
    printf("Usage: stat [filename]\n");
    exit();
  }
  fd = fopen(argv[1], "r");
#endif 
  if (fd == 0) {
     printf("Error in opening file\n");
     exit();
   }

  while (!feof(fd)) {
    fscanf(fd,"%d %d %lf %lf\n", &fnum, &fsize, &rate, &time);
    timeslot += time;
    framecount++;

    if (timeslot >= 1.0000) {       /* ONE TIME SLOT HAS ELAPSED */
       rem = timeslot - 1.0000;     /* we may send partial frames */ 
       if (rem > 0) {               /* But we don't count incomplete */
          framecount--;             /*   frames as having been sent */
       }
       delta += framecount - FPS;       /* Overflow or Underflow slot? */

#ifdef DEBUG
printf("-----------------\n");
printf("timeslot %.4f\n", timeslot); 
printf("num %d, size %d, rate %.4f, time %.4f\n", fnum,fsize,rate,time);
printf("time slot passed: rem %f, delta %d, fcount %d\n", rem, 
       delta, framecount);
printf("-----------------\n");
#endif

      if (delta < maxUflow) {          /* delta < 0 => underflow */
        maxUflow = delta;              /* store largest underflow */
      }
      if (delta > maxOflow) {          /* delta > 0 => overflow */
        maxOflow = delta;              /* store largest overflow */
      }

      if (framecount > FPS){
        OflowSlots++;
      }
      if (framecount < FPS){
        UflowSlots++;
      }
      cnt = UflowSlots - OflowSlots;
      if (cnt > maxUflowSlots) {
        maxUflowSlots = cnt;
      }
      cnt = OflowSlots - UflowSlots;
      if (cnt > maxOflowSlots) {
        maxOflowSlots = cnt;
      }
      timeslot = rem;
      if (rem > 0) framecount = 1;
      else framecount = 0;
    }
  }

#ifndef NOTSELFRUN
  printf("\n");
  printf("OVERFLOW/UNDERFLOW-SLOT DATA:\n");
  printf("Overflow Slots       : %d\n", maxOflowSlots);
  printf("Underflow Slots      : %d\n", maxUflowSlots);
  printf("Max. Underflow Frames: %d\n", maxUflow);
  printf("Max. Overflow Frames : %d\n", maxOflow);
  printf("\n");
#endif

  fseek(fd,0,0);
  buffsize = 0;
  for(i=0; i > maxUflow; i--) {    /* maxUflow < i < 0 */ 
    fscanf(fd,"%d %d %lf %lf\n", &fnum, &fsize, &rate, &time);
    buffsize += fsize;
  }

#ifdef NOTSELFRUN
                                  /* returns MAXIMUM buffer requirment */
  fprintf(fdo,"%d  %d  %.3lf\n", m, buffsize, computeOflow(fd, maxUflow)); 
  fclose(fdo);
#else 
  printf("Underflow Buffer requirement (bits): %d\n", buffsize); 
  printf("Maximum Buffer requirement (bits) : %.3lf\n", 
         computeOflow(fd, maxUflow));
  printf("\n\n");
#endif 

  fclose(fd);
}

/*-----------------------------------------*/

double computeOflow(fd, Uflow)
  FILE *fd;
  int Uflow;     /* max underflow frames */

{  
  struct Queue *tmpPtr = NULL;
  struct Queue *Qptr = NULL;
  int fno, fsize;
  int Framecount;
  double TimeCount = 0;
  double rem;
  int i, tmp;
  double maxOflow=0;
  double tmpOflow;
  int slotno = 0;
  double rate, time;
  char startDisplay = 0;
  int slotOverflow;
  int slotOverflowTotal;
  int framesBuffered;

  fseek(fd,0,0);

  Framecount = 0;
  Uflow *= -1;

  TimeCount = 0;
  slotOverflowTotal = 0;
  framesBuffered = 0;
  while (!feof(fd)) {                     
     fscanf(fd, "%d %d %lf %lf\n", &fno, &fsize, &rate, &time);
     Framecount++;
     TimeCount += time;

     Qptr = (struct Queue *)malloc(sizeof(struct Queue));
     Qptr->fsize = fsize;
     Qptr->fno = fno;
     Qptr->next = NULL;

     if (Qhead == NULL) {
        Qhead = Qptr;
        Consumed = Qptr;
        Current = Qptr;
     } else {
         Current->next = Qptr;
         Current = Qptr;
     } 
     
     if (TimeCount > 1.0000) {      /* one time slot has elapsed */
        rem = TimeCount - 1.0000;
        if (rem > 0) {
            Framecount--;
        }
        /* 
         * Check if ready to start display consumption. If so,
         * Measure Qlength and consume a slot.
         */
        if (startDisplay) {
           if ((tmpOflow = getOverflow(maxOflow, rate, rem)) > maxOflow) {
              maxOflow = tmpOflow; 
	    }
           if (consumeSlot() < 0) {
              printf("Warning!! Underflow Condition\n");
              return(-125);
           }
        } 
        /*
         * Check if Uflow has been filled. If so, start display
         * after appropriate slots have been buffered. We must
         * buffer the overflow requirement, PLUS the number of
         * frames in the buffered slots that contributed to the
         * overflow factor in underflow computation.  
         */
        if (!startDisplay) {
           framesBuffered += Framecount;
           slotOverflow = Framecount - FPS;
           if (slotOverflow > 0) {
              slotOverflowTotal += slotOverflow;
	    }
           if (framesBuffered >= Uflow) {  
              if ((Uflow + slotOverflowTotal) <= framesBuffered) {
                  startDisplay = 1; 
	      }
           } 
        }
        /*
         * Set-up for next time slot.
         * Account for previous slot spill over.
         */
        TimeCount = rem;
        if (rem > 0) Framecount = 1;
        else Framecount = 0;
     } 
   }  /* end while */ 

   while (Qhead != NULL) {  /* Empty up the queue */
      consumeSlot();
   }
   return(maxOflow); 
}

/*--------------------------*/
double getOverflow(oflow, rate, rem)
   double oflow, rate, rem;
{
   struct Queue *Qptr;
   double over = 0;

   Qptr = Qhead;
   if (Qhead == NULL) {
      return 0;
    } else {
       while (Qptr != NULL) {
          over += Qptr->fsize * 1.0;           
          Qptr = Qptr->next;
       }
       over -= rate * rem;  /* substract bits to be pushed over */
                            /* to the next time slot.           */
       return over;
    }
}


/*--------------------------*/
int consumeSlot() {
   int i;
   struct Queue *Qptr;
 
   i = 0;
   while (i < FPS) {
      if (Qhead == NULL) { 
         return -1;
      }
      Qptr = Qhead;
      Qhead = Qptr->next;
      free(Qptr);
      ConsumedFrames++;
      i++;
    }
} /* end consumeSlot() */ 

















